/*
 * Copyright (c) 2004-2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/**
 ** zoneCmdTaskCodels.c
 **
 ** Codels appeles par la tache zoneCmdTask
 **
 ** Author: Sara Fleury
 ** Date: 9/15/01
 **
 **/
 
#ifndef UNIX
#include <vxWorks.h>
#else
#include <portLib.h>
#include <math.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

#include <h2mathLib.h>

#include "server/zoneHeader.h"
#include "server/zonePosterWriteLibProto.h"
#include "zoneStruct.h"
#include "zoneConst.h"

#include "libio/zoneIO.h"
#include "zoneInt.h"

#include <server/seglocPosterLib.h>

#include "zoneNavMap.h"
#include "pomStruct.h"
#include "polygons.h"


/* internal data */
static ZONE_MOBILE_STATE mobiles[ZONE_NBMAX_MOBILES];
static ZONE_TYPE_DESC zoneStructs[ZONE_NB_TYPES];
static int zoneNextUniqueId=1;

/**** Initialization flags
 The module is operational if all of these flags are TRUE
 */
/* TRUE if there is at least one zone file by zone type */
static BOOL zoneLoadEnvDone = FALSE;

/* -- PRIVATE FUNCTIONS --------------------------------------------- */

static int zoneMainMonitoring(ZONE_MOBILE_ID mobileId);

static int zonePolyChangeState(ZONE_MOBILE_ID mobileId,
			       ZONE_STATE finalState, 
			       ZONE_TYPE *type, 
			       int *zoneIndex);

static void  zoneGrowPoly(ZONE_POINT dst[], ZONES_MAP *src, int index);

static BOOL lookForZoneFromName(char *zoneName, 
				ZONE_TYPE *type, int *zoneInd);
static BOOL lookForZoneFromId(int zoneId, 
			      ZONE_TYPE *type, int *zoneInd);

static BOOL zoneLoadEnvModels(ZONE_INIT *initParams, int *bilan);

static BOOL getOutPoint (double point[2], double *seg, 
			 double *prevPt, double *nextPt,
			 double requiredDistToSeg,
			 double outPt[2]);

static BOOL zoneGetOut (ZONE_GET_OUT_FROM_OBST *getOut,
			ZONE_POINT *outPoint);

static BOOL zoneFindPoster(char *name, POSTER_ID *posterId, 
			   ZONE_POSTER_POS_TYPE *posterStructType) ;
static BOOL zoneFindPosPoster(int *bilan);
static BOOL zoneReady (int* bilan) ;

/*------------------------------------------------------------------------
 * Start
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_zone_CANNOT_READ_POSTER
 */

/* zoneStartExec  -  codel EXEC de Start
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneStartExec (ZONE_INIT *initParams, int *bilan)
{
  ZONE_MOBILE_DESC mobileParams;
  ZONE_MOBILE_ID mobileId;
  int i;

  zoneLoadEnvDone = FALSE;

  zoneSetRobotDiameter(initParams->robotDiameter);

  if (!zoneLoadEnvModels(initParams, bilan))
    return ETHER;

  /* look for pos poster */
  strcpy(mobileParams.posPosterName.name, initParams->posPosterName.name);
  for (i=0; i<ZONE_NB_TYPES; i++) {
    mobileParams.monitorZoneType[i] =
      mobiles[ZONE_MOBILE_ROBOT].monitorZoneType[i];
  }
  zoneMobileToMonitorExec(&mobileParams, &mobileId, bilan);
/*   zoneFindPosPoster(&bilanTmp); */
  SDI_F->obstState = -1;

  return ETHER;
}

/*----------------------------------------------------------------------------
 * zonePointsListFirstPt  -  renvoie l'firstPt d'une zone dans le tableau 
 *                          de points a partir de l'id de la zone 
 */
static int zonePointsListFirstPt(int id, ZONES_MAP *zonesMap){
  int res=0,i;
  if ( id < 0 || id >= zonesMap->nbZones )
    return -1;
  else {
    for(i=0;i<id;i++)
      res += zonesMap->nbPtsPerZone[i];
    return res;
  }
}


/*------------------------------------------------------------------------
 *
 * zoneCmdTaskInit  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS zoneCmdTaskInit(int *bilan)
{
  int i, j, m;
  ZONE_TYPE_DESC *zone;

  SDI_F->initParams.simulFlag = ZONE_TRUE;
  strcpy(SDI_F->initParams.posPosterName.name, ZONE_DEFAULT_POSTER_POS);
  strcpy(SDI_F->initParams.zonePathFiles.name, ZONE_DEFAULT_ZONES);
  /*strcpy(SDI_F->zoneFile.fileName.name, ZONE_DEFAULT_POSTER_POS);*/

  /* init mobiles */
  for (m=0; m<ZONE_NBMAX_MOBILES; m++) {
    mobiles[m].pomPosterId = NULL;
    mobiles[m].lastEnterEventZoneIndex = -1; 
    mobiles[m].lastLeaveEventZoneIndex = -1; 
  }

  for (i=0; i<ZONE_NB_TYPES; i++) {
    /* by default monitor all zones for ZONE_MOBILE_ROBOT */
    mobiles[ZONE_MOBILE_ROBOT].monitorZoneType[i] = ZONE_TRUE;
    for (j=0; j<ZONE_NBMAX_ZONES; j++)
      bzero(zoneStructs[i].zonesMap.desc[j].name.name, sizeof(GEN_STRING32));
  }

  /* by default monitor no zones for other mobiles */
  for (i=ZONE_MOBILE_1; i<=ZONE_MOBILE_3; i++) {
    for(j=0;j<ZONE_NB_TYPES;j++) {
      mobiles[i].monitorZoneType[j] = ZONE_FALSE;
    }
  }

  /* initialisation des posters */
  /* TARGET */
  zone = &zoneStructs[ZONE_TARGET];

  zone->posterId[ZONE_POSTER_MAP] = ZONE_TARGETMAP_POSTER_ID;
  zone->zonesMap.nbZones = 0;
  if (zoneTARGETMapZONES_MAPPosterWrite (zone->posterId[ZONE_POSTER_MAP],
					 &zone->zonesMap) != OK) {
    printf ("zone: cannot write poster %s\n", ZONE_TARGETMAP_POSTER_NAME);
    return(ERROR);
  }

  zone->posterId[ZONE_POSTER_ROB_IN] = ZONE_TARGETROBOTIN_POSTER_ID;
  zone->zonesOccupied[ZONE_MOBILE_ROBOT].nbZones = 0;
  zone->zonesOccupied[ZONE_MOBILE_ROBOT].nbOccupiedZones = 0;
  for(i=0;i<ZONE_NBMAX_ZONES;i++)
    zone->zonesOccupied[ZONE_MOBILE_ROBOT].occupiedFlag[i] = ZONE_OUTSIDE_POLY;
  if (zoneTARGETRobotInZONES_OCCUPIEDPosterWrite (zone->posterId[ZONE_POSTER_ROB_IN],
					       &zone->zonesOccupied[ZONE_MOBILE_ROBOT]) != OK) {
    printf ("zone: cannot write poster %s\n", ZONE_TARGETROBOTIN_POSTER_NAME);
    return(ERROR);
  }

  /* SPECIAL */
  zone = &zoneStructs[ZONE_SPECIAL];
  zone->posterId[ZONE_POSTER_MAP] = ZONE_SPECIALMAP_POSTER_ID;
  zone->zonesMap.nbZones = 0;
  if (zoneSPECIALMapZONES_MAPPosterWrite (zone->posterId[ZONE_POSTER_MAP],
				       &zone->zonesMap) != OK) {
    printf ("zone: cannot write poster %s\n", ZONE_SPECIALMAP_POSTER_NAME);
    return(ERROR);
  }

  zone->posterId[ZONE_POSTER_ROB_IN] = ZONE_SPECIALROBOTIN_POSTER_ID;
  zone->posterOccupiedWriteFunc[ZONE_MOBILE_ROBOT] = zoneSPECIALRobotInZONES_OCCUPIEDPosterWrite;
  zone->posterId[ZONE_POSTER_M1_IN] = ZONE_SPECIALMOBILE1IN_POSTER_ID;
  zone->posterOccupiedWriteFunc[ZONE_MOBILE_1] = zoneSPECIALMobile1InZONES_OCCUPIEDPosterWrite;
  zone->posterId[ZONE_POSTER_M2_IN] = ZONE_SPECIALMOBILE2IN_POSTER_ID;
  zone->posterOccupiedWriteFunc[ZONE_MOBILE_2] = zoneSPECIALMobile2InZONES_OCCUPIEDPosterWrite;
  zone->posterId[ZONE_POSTER_M3_IN] = ZONE_SPECIALMOBILE3IN_POSTER_ID;
  zone->posterOccupiedWriteFunc[ZONE_MOBILE_3] = zoneSPECIALMobile3InZONES_OCCUPIEDPosterWrite;
  for (j=0; j<ZONE_NBMAX_MOBILES; j++) {
    zone->zonesOccupied[j].nbZones = 0;
    zone->zonesOccupied[j].nbOccupiedZones = 0;
    
    if ((zone->posterOccupiedWriteFunc[j])(zone->posterId[ZONE_POSTER_ROB_IN+j],
					 &zone->zonesOccupied[j]) != OK) {
      printf ("zone: cannot write poster ZONE_SPECIALMOBILE%dIN\n",j);
      return(ERROR);
    }
  
    for(i=0;i<ZONE_NBMAX_ZONES;i++)
      zone->zonesOccupied[j].occupiedFlag[i] = ZONE_OUTSIDE_POLY;
  }

  /* VIRTUAL_OBST */
  zone = &zoneStructs[ZONE_VIRTUAL_OBST];

  zone->posterId[ZONE_POSTER_MAP] = ZONE_VIRTUAL_OBSTMAP_POSTER_ID;
  zone->zonesMap.nbZones = 0;

  if (zoneVIRTUAL_OBSTMapZONES_MAPPosterWrite (zone->posterId[ZONE_POSTER_MAP],
					       &zone->zonesMap) != OK) {
    printf ("zone: cannot write poster %s\n", ZONE_VIRTUAL_OBSTMAP_POSTER_NAME);
    return(ERROR);
  }

  zone->posterId[ZONE_POSTER_ROB_IN] = ZONE_VIRTUAL_OBSTROBOTIN_POSTER_ID;
  zone->zonesOccupied[ZONE_MOBILE_ROBOT].nbZones = 0;
  zone->zonesOccupied[ZONE_MOBILE_ROBOT].nbOccupiedZones = 0;
  for(i=0;i<ZONE_NBMAX_ZONES;i++)
    zone->zonesOccupied[ZONE_MOBILE_ROBOT].occupiedFlag[i] = ZONE_OUTSIDE_POLY;
  if (zoneVIRTUAL_OBSTRobotInZONES_OCCUPIEDPosterWrite (zone->posterId[ZONE_POSTER_ROB_IN],
						     &zone->zonesOccupied[ZONE_MOBILE_ROBOT]) != OK) {
    printf ("zone: cannot write poster %s\n", ZONE_VIRTUAL_OBSTROBOTIN_POSTER_NAME);
    return(ERROR);
  }
  
  printf("Posters created\n");
 

  return OK;
}


/*------------------------------------------------------------------------
 *
 * zoneLoadEnvModels
 *
 * Description: 
 * 
 * Bilans:      OK 
 * 
 * Retourne:    BOOL
 */

static char *fileNames[] =  ZONE_FILE_NAMES;

static BOOL zoneLoadEnvModels(ZONE_INIT *initParams, int *bilan)
{
  /* init default target zones */
  sprintf(SDI_F->zoneFile.fileName.name, "%s/%s",
	  initParams->zonePathFiles.name, fileNames[ZONE_TARGET]);
  SDI_F->zoneFile.type = ZONE_TARGET;
  zoneReadFileExec(&SDI_F->zoneFile, bilan);


  /* init default special zones */
  sprintf(SDI_F->zoneFile.fileName.name, "%s/%s",
	  initParams->zonePathFiles.name, fileNames[ZONE_SPECIAL]);
  SDI_F->zoneFile.type = ZONE_SPECIAL;
  zoneReadFileExec(&SDI_F->zoneFile, bilan);


  /* init default obst zones */
  sprintf(SDI_F->zoneFile.fileName.name, "%s/%s",
	  initParams->zonePathFiles.name, fileNames[ZONE_VIRTUAL_OBST]);
  SDI_F->zoneFile.type = ZONE_VIRTUAL_OBST;
  zoneReadFileExec(&SDI_F->zoneFile, bilan);

  zoneLoadEnvDone = TRUE;

  return TRUE;
}

static BOOL zoneFindPosPoster(int *bilan)
{
  /* Init  the pom poster */
  if (!SDI_F->initParams.simulFlag && 
      mobiles[ZONE_MOBILE_ROBOT].pomPosterId == NULL) {
    
    /* look for its pos poster */
    if (!zoneFindPoster(SDI_F->initParams.posPosterName.name,
			&mobiles[ZONE_MOBILE_ROBOT].pomPosterId,
			&mobiles[ZONE_MOBILE_ROBOT].pomPosterStructType)) {
      mobiles[ZONE_MOBILE_ROBOT].pomPosterId = NULL;
      *bilan = S_zone_NOT_CONNECTED_TO_POM;
      return FALSE;
    }
  }
  return TRUE;
}


static BOOL zoneReady (int* bilan) 
{ 
  /* test if env known */
  if (!zoneLoadEnvDone) {
    *bilan = S_zone_NOT_READY;
    return FALSE;
  }

  /* look for pos poster */
  if (mobiles[ZONE_MOBILE_ROBOT].pomPosterId==NULL) {
    if (!zoneFindPosPoster(bilan)) {
      *bilan = S_zone_NOT_READY;
      return FALSE;
    }
  }
  return TRUE;
}


/*------------------------------------------------------------------------
 *
 * zoneCmdTaskPerm  --  codel de l'activite' permanente
 *
 * Description: 
 * 
 * Bilans:      OK 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS zoneCmdTaskPerm(int *bilan)
{
  int i;

  if (!zoneReady(bilan))
      return OK;

  zoneMainMonitoring(ZONE_MOBILE_ROBOT);

  zoneMainMonitoring(ZONE_MOBILE_1);
  zoneMainMonitoring(ZONE_MOBILE_2);
  zoneMainMonitoring(ZONE_MOBILE_3);

  /* update states posters for ZONE_MOBILE_ROBOT and all ZONE_SPECIAL */
  if (zoneTARGETRobotInZONES_OCCUPIEDPosterWrite(zoneStructs[ZONE_TARGET].posterId[ZONE_POSTER_ROB_IN],
					      &zoneStructs[ZONE_TARGET].zonesOccupied[ZONE_MOBILE_ROBOT]) != OK) {
    printf ("zone: cannot write poster TARGET ZONE_TARGETROBOTIN\n");
    *bilan = S_zone_CANNOT_WRITE_POSTER;
  }

  for (i=0; i<ZONE_NBMAX_MOBILES; i++) {
    if (mobiles[i].monitorZoneType[ZONE_SPECIAL]) {
      if (zoneSPECIALRobotInZONES_OCCUPIEDPosterWrite(zoneStructs[ZONE_SPECIAL].posterId[i+ZONE_POSTER_ROB_IN],
						   &zoneStructs[ZONE_SPECIAL].zonesOccupied[i]) != OK) {
	printf ("zone: cannot write poster SPECIAL ZoneOccupied mobile No %d\n", i);
	*bilan = S_zone_CANNOT_WRITE_POSTER;
      }
    }
  }

  if (zoneVIRTUAL_OBSTRobotInZONES_OCCUPIEDPosterWrite(zoneStructs[ZONE_VIRTUAL_OBST].posterId[ZONE_POSTER_ROB_IN],
					      &zoneStructs[ZONE_VIRTUAL_OBST].zonesOccupied[ZONE_MOBILE_ROBOT]) != OK) {
    printf ("zone: cannot write poster VIRTUAL_OBST ZoneStates\n");
    *bilan = S_zone_CANNOT_WRITE_POSTER;
  }

  return OK;
}


/* ----------------------------------------------------------------------
 * 
 * zoneMainMonitoring -
 *
 */
static int zoneMainMonitoring(ZONE_MOBILE_ID mobileId)
{
  int i, j;
  ZONE_POINT *points;
  static POM_ME_POS pomMePos;
  static POM_SENSOR_POS pomSensorPos;
  static POM_EULER_POS pomEulerPos;

  if (mobiles[mobileId].pomPosterId == NULL) {
/*       (!SDI_F->initParams.simulFlag || mobileId != ZONE_MOBILE_ROBOT)) */
    return FALSE;
  }

  /* Get the current position estimate */
  if (mobiles[mobileId].pomPosterId) {
    switch(mobiles[mobileId].pomPosterStructType) {
    case ZONE_POSTER_POM_POS:
      if(zonePOM_POSPosterRead(mobiles[mobileId].pomPosterId, 
			       &mobiles[mobileId].pomPos) != 0) {
	printf ("zone : Can't get pomPos for mobile %d ! It is removed.\n", mobileId);
	mobiles[mobileId].pomPosterId = NULL;
	return FALSE;
      }
      break;
    case ZONE_POSTER_POM_ME_POS:
      if(zonePOM_ME_POSPosterRead(mobiles[mobileId].pomPosterId, 
			       &pomMePos) != 0) {
	printf ("zone : Can't get pomPos for mobile %d ! It is removed.\n", mobileId);
	mobiles[mobileId].pomPosterId = NULL;
	return FALSE;
      }
      mobiles[mobileId].pomPos.date = pomMePos.date1;
      mobiles[mobileId].pomPos.mainToOrigin = pomMePos.main;
      mobiles[mobileId].pomPos.mainToBase = pomMePos.main;
      mobiles[mobileId].pomPos.VLocal = pomMePos.VLocal;
      break;
    case ZONE_POSTER_POM_SENSOR_POS:
      if(zonePOM_SENSOR_POSPosterRead(mobiles[mobileId].pomPosterId, 
				      &pomSensorPos) != 0) {
	printf ("zone : Can't get pomPos for mobile %d ! It is removed.\n", mobileId);
	mobiles[mobileId].pomPosterId = NULL;
	return FALSE;
      }
      /* XXXXX Shouldn't we use sensorToMain ???? */
      mobiles[mobileId].pomPos.date = pomSensorPos.date;
      mobiles[mobileId].pomPos.mainToOrigin = pomSensorPos.mainToOrigin;
      mobiles[mobileId].pomPos.mainToBase = pomSensorPos.mainToBase;
      mobiles[mobileId].pomPos.VLocal = pomSensorPos.VLocal;
      break;
    case ZONE_POSTER_POM_EULER_POS:
      if(zonePOM_EULER_POSPosterRead(mobiles[mobileId].pomPosterId, 
				      &pomEulerPos) != 0) {
	printf ("zone : Can't get pomPos for mobile %d ! It is removed.\n", mobileId);
	mobiles[mobileId].pomPosterId = NULL;
	return FALSE;
      }
      mobiles[mobileId].pomPos.date = pomEulerPos.date;
      mobiles[mobileId].pomPos.mainToOrigin.euler = pomEulerPos.main;
      mobiles[mobileId].pomPos.mainToBase.euler = pomEulerPos.main;
      mobiles[mobileId].pomPos.VLocal.euler = pomEulerPos.VLocal;
      break;
    }
  }

  /* MainToOrigin is the best pose */
  mobiles[mobileId].currentConfig.x = mobiles[mobileId].pomPos.mainToOrigin.euler.x;
  mobiles[mobileId].currentConfig.y = mobiles[mobileId].pomPos.mainToOrigin.euler.y;
  mobiles[mobileId].currentConfig.theta = mobiles[mobileId].pomPos.mainToOrigin.euler.yaw;
  
  /* Test if mobile is inside poly */
  /* for each type of zone */
  for(j=0;j<ZONE_NB_TYPES;j++) {

    zoneStructs[j].zonesOccupied[mobileId].nbOccupiedZones = 0;

    /* don't mind about this zone */
    if (!mobiles[mobileId].monitorZoneType[j])
      continue;

    for (i=0; i<zoneStructs[j].zonesMap.nbZones; i++) {
      int firstPt = zonePointsListFirstPt(i, &zoneStructs[j].zonesMap);
      if (zoneStructs[j].zonesMap.nbPtsPerZone[i] <= 1)
	continue;
      
      /* Select poly according to type of the zone */
      if (zoneStructs[j].zonesMap.desc[i].class == ZONE_TOPO)
	points = zoneStructs[j].zonesMap.points + firstPt;
      else
	points = zoneStructs[j].growedPoints + firstPt;
      
      /* INSIDE THE ZONE */
      if (PLpointInPolygon((double *) &mobiles[mobileId].currentConfig,
	zoneStructs[j].zonesMap.nbPtsPerZone[i],
	(double *)points)) {
	/* Test if state changed */
	if (zoneStructs[j].zonesOccupied[mobileId].occupiedFlag[i] != ZONE_INSIDE_POLY) {
	  
	  printf ("Mobile %d ENTER %s (id %d type %d)\n", 
		  mobileId,
		  zoneStructs[j].zonesMap.desc[i].name.name,
		  zoneStructs[j].zonesMap.desc[i].id, 
		  zoneStructs[j].zonesMap.desc[i].class);
	  mobiles[mobileId].lastEnterEventZoneIndex = i;
	  mobiles[mobileId].lastEnterEventZoneType = j;
	}
	/* Record new state */
	zoneStructs[j].zonesOccupied[mobileId].occupiedFlag[i] = ZONE_INSIDE_POLY;

	/* record zone */
	strcpy(zoneStructs[j].zonesOccupied[mobileId].occupiedList[zoneStructs[j].zonesOccupied[mobileId].nbOccupiedZones].name, 
	       zoneStructs[j].zonesMap.desc[i].name.name);
	       
	zoneStructs[j].zonesOccupied[mobileId].nbOccupiedZones++;

	/* INSIDE A POLY */
	
	/* OUTSIDE THE POLY */
      } else {
	/* Outside, test if state changed */
	if (zoneStructs[j].zonesOccupied[mobileId].occupiedFlag[i] != ZONE_OUTSIDE_POLY) {
	  printf ("Mobile %d LEAVE %s (id %d type %d)\n",
		  mobileId,
		  zoneStructs[j].zonesMap.desc[i].name.name,
		  zoneStructs[j].zonesMap.desc[i].id, 
		  zoneStructs[j].zonesMap.desc[i].class);
	  mobiles[mobileId].lastLeaveEventZoneIndex = i;
	  mobiles[mobileId].lastLeaveEventZoneType = j;
	}
	/* Record new state */
	zoneStructs[j].zonesOccupied[mobileId].occupiedFlag[i] = ZONE_OUTSIDE_POLY;
      } /* OUTSIDE THE POLY */
      
    } /* FOR EACH ZONE OF A GIVEN TYPE */
    
  } /* FOR EACH TYPE */

  return TRUE;
}  

/*------------------------------------------------------------------------
 * zoneGetZoneByNameExec  -  request GetZone
 *
 * Description:    
 * Report (bilan): OK 
 *                 S_zone_BAD_VALUE
 *
 * Return:    ACTIVITY_EVENT
 */

ACTIVITY_EVENT zoneGetZoneByNameExec (GEN_STRING128 *zoneName,
				      ZONE_PARAMS *zoneParams,
				      int *bilan)
{
  int zoneIndex;
  ZONE_TYPE zoneType;
  ZONE_TYPE_DESC *zone;

  if (!lookForZoneFromName(zoneName->name, &zoneType, &zoneIndex)) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }

  zone = &zoneStructs[zoneType];

  zoneParams->zoneDesc = zone->zonesMap.desc[zoneIndex];
  
  zoneParams->occupiedFlag = zone->zonesOccupied[ZONE_MOBILE_ROBOT].occupiedFlag[zoneIndex];
  zoneParams->nbPts = zone->zonesMap.nbPtsPerZone[zoneIndex];
  zoneParams->firstPt = zonePointsListFirstPt(zoneIndex, &zone->zonesMap);
  zoneParams->lastPt = zoneParams->firstPt + zoneParams->nbPts - 1;
  
  return ETHER;
}

/*------------------------------------------------------------------------
 * zoneGetZoneByIdExec  -  request GetZone
 *
 * Description:    
 * Report (bilan): OK 
 *                 S_zone_BAD_VALUE
 *
 * Return:    ACTIVITY_EVENT
 */

ACTIVITY_EVENT zoneGetZoneByIdExec (int *zoneId,
				    ZONE_PARAMS *zoneParams,
				    int *bilan)
{
  int zoneIndex;
  ZONE_TYPE zoneType;
  ZONE_TYPE_DESC *zone;

  if (!lookForZoneFromId(*zoneId, &zoneType, &zoneIndex)) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }

  zone = &zoneStructs[zoneType];

  zoneParams->zoneDesc = zone->zonesMap.desc[zoneIndex];
  
  zoneParams->occupiedFlag = zone->zonesOccupied[ZONE_MOBILE_ROBOT].occupiedFlag[zoneIndex];
  zoneParams->nbPts = zone->zonesMap.nbPtsPerZone[zoneIndex];
  zoneParams->firstPt = zonePointsListFirstPt(zoneIndex, &zone->zonesMap);
  zoneParams->lastPt = zoneParams->firstPt + zoneParams->nbPts - 1;
  
  return ETHER;
}

/*------------------------------------------------------------------------
 * MonitorEnterZoneId
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* zoneEnterPolyExec  -  codel EXEC de EnterPoly
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneMonitorEnterZoneIdExec (ZONE_MOBILE_AND_ZONE_ID *monParams,
					   ZONE_IDENTITY *identity,  int *bilan)
{
  int zoneIndex, zoneChanged;
  ZONE_TYPE zoneType;
  
  if (monParams->mobileId<0 || monParams->mobileId>=ZONE_NBMAX_MOBILES) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }

  identity->mobileId = monParams->mobileId;


  if (!zoneReady(bilan))
    return ETHER;

  if (!lookForZoneFromId(monParams->zoneId, &zoneType, &zoneIndex)) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }

  zoneChanged = zonePolyChangeState(monParams->mobileId, ZONE_INSIDE_POLY, 
				    &zoneType, &zoneIndex);
  printf ("w %d c %d\n", monParams->zoneId, zoneChanged);
  if (monParams->zoneId == zoneChanged) {
    strcpy(identity->name.name, zoneStructs[zoneType].zonesMap.desc[zoneIndex].name.name);
    identity->type = zoneType;
    identity->id = monParams->zoneId;
    return ETHER;
  }

  return EXEC;
}

/*------------------------------------------------------------------------
 * MonitorEnter
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* zoneMonitorEnterExec  -  codel EXEC de MonitorEnter
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneMonitorEnterExec (ZONE_MOBILE_ID *mobileId,
				     ZONE_IDENTITY *identity, int *bilan)
{
  int zoneId;
  int zoneIndex;
  ZONE_TYPE zoneType;

  if (*mobileId<0 || *mobileId>=ZONE_NBMAX_MOBILES) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  } 
 
  identity->mobileId = *mobileId;
 

  if (!zoneReady(bilan))
      return ETHER;

  if ((zoneId = zonePolyChangeState(*mobileId, ZONE_INSIDE_POLY, 
				    &zoneType, &zoneIndex))) {
    strcpy(identity->name.name, zoneStructs[zoneType].zonesMap.desc[zoneIndex].name.name);
    identity->type = zoneType;
    identity->id = zoneId;
    return ETHER;
  }

  return EXEC;
}



/*------------------------------------------------------------------------
 * MonitorLeaveZoneId
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* zoneLeavePolyExec  -  codel EXEC de LeavePoly
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneMonitorLeaveZoneIdExec (ZONE_MOBILE_AND_ZONE_ID *monParams, 
					   ZONE_IDENTITY *identity,
					   int *bilan)
{
  int zoneIndex;
  ZONE_TYPE zoneType;
  
  if (monParams->mobileId<0 || monParams->mobileId>=ZONE_NBMAX_MOBILES) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }

  identity->mobileId = monParams->mobileId;

  if (!zoneReady(bilan))
    return ETHER;

  if (!lookForZoneFromId(monParams->zoneId, &zoneType, &zoneIndex)) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }

  if (monParams->zoneId == zonePolyChangeState(monParams->mobileId, 
					       ZONE_OUTSIDE_POLY, 
					       &zoneType, &zoneIndex)) {
    strcpy(identity->name.name, zoneStructs[zoneType].zonesMap.desc[zoneIndex].name.name);
    identity->type = zoneType;
    identity->id = monParams->zoneId;
    return ETHER;
  }

  return EXEC;
}

/*------------------------------------------------------------------------
 * MonitorLeave
 *
 * Description: 
 *
 * Bilans:      OK 
 */


/* zoneLeavePolyExec  -  codel EXEC de LeavePoly
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneMonitorLeaveExec (ZONE_MOBILE_ID *mobileId,
				     ZONE_IDENTITY *identity, int *bilan)
{
  int zoneId;
  ZONE_TYPE zoneType;
  int zoneIndex;

  if (*mobileId<0 || *mobileId>=ZONE_NBMAX_MOBILES) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }
  identity->mobileId = *mobileId;

  if (!zoneReady(bilan))
    return ETHER;

  if ((zoneId = zonePolyChangeState(*mobileId, ZONE_OUTSIDE_POLY, 
				    &zoneType, &zoneIndex))) {
    strcpy(identity->name.name, zoneStructs[zoneType].zonesMap.desc[zoneIndex].name.name);
    identity->type = zoneType;
    identity->id = zoneId;
    return ETHER;
  }

  return EXEC;
}

/*------------------------------------------------------------------------
 * zonePolyChangeState
 *
 * Description: Test if a zone has a new state relatively to the
 *              last request.
 *        
 * Bilans:      OK 
 */


static int zonePolyChangeState(ZONE_MOBILE_ID mobileId,
			       ZONE_STATE finalState, 
			       ZONE_TYPE *type, 
			       int *zoneIndex)
{
  /* enter a zone ? */
  if (finalState == ZONE_INSIDE_POLY && mobiles[mobileId].lastEnterEventZoneIndex != -1) {
      *zoneIndex = mobiles[mobileId].lastEnterEventZoneIndex;
      *type = mobiles[mobileId].lastEnterEventZoneType;
      mobiles[mobileId].lastEnterEventZoneIndex = -1;
  }

  /* leave a zone ? */
  else if (finalState == ZONE_OUTSIDE_POLY && mobiles[mobileId].lastLeaveEventZoneIndex != -1) {
    *zoneIndex = mobiles[mobileId].lastLeaveEventZoneIndex;
    *type = mobiles[mobileId].lastLeaveEventZoneType;
    mobiles[mobileId].lastLeaveEventZoneIndex = -1;
  }

  /* no event */
  else 
    return 0;

  return zoneStructs[*type].zonesMap.desc[*zoneIndex].id;
}


/*
 * zoneGrowPoly ---------------------------------------------------------
 *
 * Make a polygon bigger (or smaller if "OUTSIDE")
 */

static void
zoneGrowPoly(ZONE_POINT dst[], ZONES_MAP *src, int index)
{
   int firstPt,nbPts;
   
   /* compute points */
   firstPt = zonePointsListFirstPt(index, src);
   nbPts = src->nbPtsPerZone[index];
   PLgrowPolygon(nbPts, 
		 (double *)&src->points[firstPt], SDI_F->initParams.robotDiameter/2.0,
		 (double *)&dst[firstPt]);
}


/*------------------------------------------------------------------------
 * GetTargetDirection
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_zone_BAD_VALUE
 */

/* zoneGetTargetDirectionExec  -  codel EXEC de GetTargetDirection
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneGetTargetDirectionExec (ZONE_MOBILE_AND_ZONE_ID *mobileAndZone,
					   ZONE_ROBOT_TO_TARGET *target, 
					   int *bilan)
{
  int zoneIndex;
  ZONE_TYPE zoneType;
  double dx, dy;
  ZONE_CONFIG *targetConfig;

  if (mobileAndZone->mobileId<0 || mobileAndZone->mobileId>=ZONE_NBMAX_MOBILES) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  } 
  target->mobileId = mobileAndZone->mobileId;

  if (!lookForZoneFromId(mobileAndZone->zoneId, &zoneType, &zoneIndex)) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }

  targetConfig = &zoneStructs[zoneType].zonesMap.desc[zoneIndex].target;

  dx = targetConfig->x - mobiles[mobileAndZone->mobileId].currentConfig.x;
  dy = targetConfig->y - mobiles[mobileAndZone->mobileId].currentConfig.y;

  target->dist = hypot(dx, dy);
  target->thetaAbs = atan2(dy, dx);
  target->thetaRel = angleLimit(target->thetaAbs - mobiles[mobileAndZone->mobileId].currentConfig.theta);

  printf ("zoneTarget: %s Rob %g %g %g Tar %g %g T-R = %g %g %g ROT %g\n",
	  zoneStructs[zoneType].zonesMap.desc[zoneIndex].name.name,
	  mobiles[mobileAndZone->mobileId].currentConfig.x, mobiles[mobileAndZone->mobileId].currentConfig.y, mobiles[mobileAndZone->mobileId].currentConfig.theta,
	  targetConfig->x, targetConfig->y,
	  dx, dy, target->thetaAbs, target->thetaRel);

  return ETHER;
}


/*------------------------------------------------------------------------
 * ReadFile
 *
 * Description: read a Zone File of a certain type and create the associated posters 
 *
 * Bilans:      OK 
 *              S_zone_CANNOT_OPEN_FILE
 */

/* zoneReadFileExec  -  codel EXEC de ReadFile
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneReadFileExec (ZONE_FILE *zoneIdent, int *bilan)
{
  int i, m;
  ZONE_TYPE_DESC *zone;

  zone = &zoneStructs[zoneIdent->type];
  
  strncpy(zone->fileName.name, zoneIdent->fileName.name, sizeof(GEN_STRING128));
  if (!zoneReadFile(zone->fileName.name, &zone->zonesMap)) {
    printf("zoneReadFileExec: cannot open file %s\n", zone->fileName.name);
    *bilan = S_zone_CANNOT_READ_FILE;
    return ETHER;
  } 
  
  printf("zoneReadFileExec: %d nbZones in zonesDesc\n",
	 zone->zonesMap.nbZones);
  printf("zoneReadFileExec : first zone name : %s\n",
	 zone->zonesMap.desc[0].name.name);
  
  for (m=0; m<ZONE_NBMAX_MOBILES; m++)
    zone->zonesOccupied[m].nbZones = zone->zonesMap.nbZones;
  for (i=0; i<zone->zonesMap.nbZones; i++) {
    zone->zonesMap.desc[i].id = zoneNextUniqueId++;
    zoneGrowPoly(zone->growedPoints, &zone->zonesMap, i);
  }

  /* mise a jour poster */
  switch(zoneIdent->type) {
  case ZONE_TARGET:
    
    if (zoneTARGETMapZONES_MAPPosterWrite(zone->posterId[ZONE_POSTER_MAP],
					  &zone->zonesMap) != OK) {
      printf ("zone: cannot write poster TARGET ZoneMap\n");
      *bilan = S_zone_CANNOT_WRITE_POSTER;
    }
    if (zoneTARGETRobotInZONES_OCCUPIEDPosterWrite(zone->posterId[ZONE_POSTER_ROB_IN],
						&zone->zonesOccupied[ZONE_MOBILE_ROBOT]) 
	!= OK) {
      printf ("zone: cannot write poster TARGET ZoneStates\n");
      *bilan = S_zone_CANNOT_WRITE_POSTER;
    }
    break;
    
  case ZONE_SPECIAL:
    if (zoneSPECIALMapZONES_MAPPosterWrite(zone->posterId[ZONE_POSTER_MAP],
					&zone->zonesMap) != OK) {
      printf ("zone: cannot write poster SPECIAL ZoneMap\n");
      *bilan = S_zone_CANNOT_WRITE_POSTER;
    }
    if (zoneSPECIALRobotInZONES_OCCUPIEDPosterWrite(zone->posterId[ZONE_POSTER_ROB_IN],
					      &zone->zonesOccupied[ZONE_MOBILE_ROBOT]) 
	!= OK) {
      printf ("zone: cannot write poster SPECIAL ZoneStates\n");
      *bilan = S_zone_CANNOT_WRITE_POSTER;
    }
    break;

  case ZONE_VIRTUAL_OBST:
    if (zoneVIRTUAL_OBSTMapZONES_MAPPosterWrite(zone->posterId[ZONE_POSTER_MAP],
						&zone->zonesMap) 
	!= OK) {
      printf ("zone: cannot write poster VIRTUAL_OBST ZoneMap\n");
      *bilan = S_zone_CANNOT_WRITE_POSTER;
    }
    if (zoneVIRTUAL_OBSTRobotInZONES_OCCUPIEDPosterWrite(zone->posterId[ZONE_POSTER_ROB_IN],
						      &zone->zonesOccupied[ZONE_MOBILE_ROBOT]) != OK) {
      printf ("zone: cannot write poster VIRTUAL_OBST ZoneStates\n");
      *bilan = S_zone_CANNOT_WRITE_POSTER;
    }
    break;
  }

  return ETHER;
}


/*------------------------------------------------------------------------
 * SaveFile
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_zone_CANNOT_OPEN_FILE
 */

/* zoneSaveFileExec  -  codel EXEC de SaveFile
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneSaveFileExec (ZONE_TYPE *zoneType, int *bilan)
{
  char fileEnlarged[1024];
  strcpy(fileEnlarged, zoneStructs[*zoneType].fileName.name);
  strcat(fileEnlarged, "large");
  
  if (!zoneSaveFile(zoneStructs[*zoneType].fileName.name,
		    &zoneStructs[*zoneType].zonesMap,
		    zoneStructs[*zoneType].zonesMap.points)) {
    *bilan = S_zone_CANNOT_OPEN_FILE;
  }
  
  if (!zoneSaveFile(fileEnlarged,
		    &zoneStructs[*zoneType].zonesMap,
		    zoneStructs[*zoneType].growedPoints)) {
    *bilan = S_zone_CANNOT_OPEN_FILE;
  }
  return ETHER;
}


/*------------------------------------------------------------------------
 * BuildNavMap
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_zone_CANNOT_READ_POSTER
 *              S_zone_CANNOT_READ_FILE
 */

/* zoneBuildNavMapExec  -  codel EXEC de BuildNavMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneBuildNavMapExec (ZONE_SEGLOC_MAP *whereSeglocMap, int *bilan)
{
  zoneExportNavMap(whereSeglocMap, 
		   &zoneStructs[ZONE_VIRTUAL_OBST].zonesMap, 
		   bilan);
  return ETHER;
}



/*------------------------------------------------------------------------
 * SaveNavMap
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_zone_CANNOT_READ_POSTER
 *              S_zone_CANNOT_READ_FILE
 */

/* zoneSaveNavMapExec  -  codel EXEC de SaveNavMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneSaveNavMapExec (GEN_STRING128 *fileName, int *bilan)
{
  if (!zoneSaveNavMap(fileName->name)) 
    *bilan = S_zone_CANNOT_OPEN_FILE;

  return ETHER;
}

/*------------------------------------------------------------------------
 * lookForZoneFromName
 *
 *
 * Return:    
 */

static BOOL lookForZoneFromName(char *zoneName, ZONE_TYPE *type, int *zoneInd)
{
  ZONE_TYPE_DESC *zone;
  int zoneIndex, i;

  /* for each type of zone */
  for (i = 0; i < ZONE_NB_TYPES; i++) {

    zone = &zoneStructs[i];

    /* Look if zone already exists */
    for (zoneIndex=0; zoneIndex<zone->zonesMap.nbZones; zoneIndex++) {
      
      /* zoneId found */
      if (!strncmp(zoneName,
		   zone->zonesMap.desc[zoneIndex].name.name,
		   sizeof(GEN_STRING32))) {
	
	*type = i;
	*zoneInd = zoneIndex;
	return TRUE;
      }
    }  
  }
  
  return FALSE;
}

/*------------------------------------------------------------------------
 * lookForZoneFromId
 *
 *
 * Return:    
 */

static BOOL lookForZoneFromId(int zoneId, ZONE_TYPE *type, int *zoneInd)
{
  ZONE_TYPE_DESC *zone;
  int zoneIndex, i;

  /* for each type of zone */
  for (i = 0; i < ZONE_NB_TYPES; i++) {

    zone = &zoneStructs[i];

    /* Look if zone already exists */
    for (zoneIndex=0; zoneIndex<zone->zonesMap.nbZones; zoneIndex++) {
      
      /* zoneId found */
      if (zone->zonesMap.desc[zoneIndex].id == zoneId) {
	
	*type = i;
	*zoneInd = zoneIndex;
	return TRUE;
      }
    }  
  }
  
  return FALSE;
}

/*------------------------------------------------------------------------
 * SimulPos
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* zoneSimulPosExec  -  codel EXEC de SimulPos
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneSimulPosExec (ZONE_CONFIG *simulPos, int *bilan)
{

  if (!SDI_F->initParams.simulFlag) 
    printf ("zone: Do not forget to call zoneStart request for monitoring on real pos\n");

  SDI_F->initParams.simulFlag = TRUE;
  mobiles[ZONE_MOBILE_ROBOT].currentConfig = *simulPos;
  return ETHER;
}


/*------------------------------------------------------------------------
 * ExportObstsInNavZone
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_zone_CANNOT_READ_POSTER
 */

/* zoneExportObstsInNavZoneExec  -  codel EXEC de ExportObstsInNavZone
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneExportObstsInNavZoneExec (ZONE_OBSTS_FIELD *obstsField, 
					     int *bilan)
{
  if (!zoneObstsInNav(obstsField, bilan))
    return ETHER;

  return EXEC;
}

static BOOL monitorObstOn=FALSE;
static ZONE_OBST_STATE prevObstState=-1;
/*------------------------------------------------------------------------
 * MonitorObsts
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* zoneMonitorObstsStart  -  codel START de MonitorObsts
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneMonitorObstsStart (ZONE_OBST_STATE *obstState, int *bilan)
{
  /* already a monitor on */
  if (monitorObstOn)
    return START;

  monitorObstOn = TRUE;

  /* test new obstState */
  printf ("MONITOR started\n");
  return zoneMonitorObstsExec(obstState, bilan);
}

/* zoneMonitorObstsExec  -  codel EXEC de MonitorObsts
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneMonitorObstsExec (ZONE_OBST_STATE *obstState, int *bilan)
{
  /* obst state change */
  if (prevObstState != SDI_F->obstState) {
    *obstState = SDI_F->obstState;
    prevObstState = *obstState;
    monitorObstOn = FALSE;
    printf ("MONITOR %s\n",
	    *obstState == ZONE_FREE ? "FREE" :
	    *obstState == ZONE_HIGH_CLUTTERED ? "HIGH" :
	    *obstState == ZONE_LOW_CLUTTERED ? "LOW" : "BLOCK");

    return ETHER;
  }

  /* still monitor */
  return EXEC;
}

/* zoneMonitorObstsInter  -  codel INTER de MonitorObsts
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneMonitorObstsInter (ZONE_OBST_STATE *obstState, int *bilan)
{
  printf ("MONITOR stopped\n");
  return ETHER;
}




/*------------------------------------------------------------------------
 * GetOut
 *
 * Description: If the position is within a virtual obstacle, computs the nearest segment of this obstacle and a position in the other side of this segment at the given distance. 
 *
 * Bilans:      OK 
 *              S_zone_NOT_IN_OBST
 */

/* zoneGetOutExec  -  codel EXEC de GetOut
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneGetOutExec (ZONE_GET_OUT_FROM_OBST *getOut,
			       ZONE_POINT *outPoint, int *bilan)
{

  /* init outpoint (default value) */
  *outPoint = getOut->pos;

  printf ("zoneGetOutExec: get out pt %5.2f %5.2f\n", getOut->pos.x, getOut->pos.y);

  /* get out a first time */
  if (!zoneGetOut(getOut, outPoint)) {
    *bilan = S_zone_NOT_IN_OBST;
    return ETHER;
  }

  /* was in an obst */
  printf ("zoneGetOutExec: new pt %5.2f %5.2f\n", outPoint->x, outPoint->y);

  /* try a second time with this new position because there's may be an pother one 
     (in case of corner) */ 
  getOut->pos = *outPoint;

  /* was again in an obst */
  if (zoneGetOut(getOut, outPoint))
    printf ("zoneGetOutExec: new pt AGAIN %5.2f %5.2f\n", outPoint->x, outPoint->y);

  return ETHER;
}



static BOOL zoneGetOut (ZONE_GET_OUT_FROM_OBST *getOut,
			ZONE_POINT *outPoint)
{
  int i, nearestSeg;
  ZONE_POINT *points;
  double minDist;
  BOOL inObst = FALSE;
  ZONE_TYPE_DESC *zoneVirtualObsts = &zoneStructs[ZONE_VIRTUAL_OBST];
  ZONES_MAP *virtualObsts = &zoneVirtualObsts->zonesMap;
  int next, prev;

  /* for each virtual obst */
  for (i=0; i<zoneVirtualObsts->zonesMap.nbZones; i++) {
    int firstPt = zonePointsListFirstPt(i, virtualObsts);
					  
    if (virtualObsts->nbPtsPerZone[i] <= 1)
      continue;
      
    /* Select poly according to type of the zone */
    if (zoneVirtualObsts->zonesMap.desc[i].class == ZONE_TOPO)
      points = virtualObsts->points + firstPt;
    else
      points = zoneVirtualObsts->growedPoints + firstPt;
      
    /* the point is within a virtual obst */
    if (PLpointInPolygon ((double *)&getOut->pos,
			  virtualObsts->nbPtsPerZone[i],
			  (double *)points)) {

      printf("zoneGetOutExec: point in %s\n",
	     zoneVirtualObsts->zonesMap.desc[i].name.name);

      /* get nearest edge */
      nearestSeg = PLpointNearestPolygonEdge ((double *)&getOut->pos,
					   virtualObsts->nbPtsPerZone[i],
					   (double *)points, &minDist);

      /* computes position out of there : IO.P1P2 = 0 && IO = minDist */
      if (nearestSeg == 0) prev = virtualObsts->nbPtsPerZone[i]-1;
      else prev = nearestSeg - 1;
      if (nearestSeg == virtualObsts->nbPtsPerZone[i]-1) next = 0;
      else next = nearestSeg + 2;

#if DEBUG>=2
      printf ("zoneGetOutExec: near edges (%5.2f %5.2f) (%5.2f %5.2f) (%5.2f %5.2f)\n", 
	     points[prev].x, points[prev].y,
	     points[nearestSeg].x, points[nearestSeg].y,
	     points[next].x, points[next].y);
#endif

      getOutPoint ((double *)&getOut->pos, (double *)&points[nearestSeg],
		   (double *)&points[prev], (double *)&points[next],
		   getOut->minDist, (double *)outPoint);

      inObst = TRUE;

      break;
    }    
  } /* each obst */
  return inObst;
}


/*----------------------------------------------------------------------
 * getOutPoint
 *
 * Description : Computes a point that is on the other side of a segment 
 *               at a given distance.
 *               If the proj of the pt is not on the seg then
 *               computes a pt on the bissector of this seg and
 *               the prev or next seg according to nearest vertice.
 */

static int getOutPoint (double point[2], double *nearestSeg, 
			double *prevPt, double *nextPt,
			double requiredDistToSeg,
			double outPt[2])
{
  int ext;

  /* proj point in seg */
  ext = PLprojPointOnSeg (point, nearestSeg, 
			  requiredDistToSeg, outPt);

  switch (ext) {
    /* proj(P) on S is in [P1,P2] */
  case 0:
    return 1;

  /* proj(P) on S is not in [P1,P2] */
  /* ext 1 nearestSeg */
  case 1:
    /* on bissector of [segPrev, nearestSeg] */
    return PLbissectorPt(prevPt, &nearestSeg[0], &nearestSeg[2], 
			 requiredDistToSeg, NULL, outPt);

  /* ext 2 nearestSeg */
  case 2:
    /* on bissector of [nearestSeg, segNext] */
    return PLbissectorPt(&nearestSeg[0], &nearestSeg[2], nextPt,
			 requiredDistToSeg, NULL, outPt);

  default:
    return 0;
  }
}


/*------------------------------------------------------------------------
 * GetCurrentTargetZone
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* zoneGetCurrentTargetZoneExec  -  codel EXEC de GetCurrentTargetZone
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneGetCurrentTargetZoneExec (ZONE_MOBILE_AND_ZONE_TYPE *mobileAndZone,
					     ZONE_IDENTITY *identity, int *bilan)
{
  int i;
  ZONE_TYPE_DESC *zone;

  if (mobileAndZone->mobileId<0 || mobileAndZone->mobileId>=ZONE_NBMAX_MOBILES) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  } 

  identity->mobileId = mobileAndZone->mobileId;

  /* TARGET */
  zone = &zoneStructs[ZONE_TARGET];

  strcpy(identity->name.name, "");
  identity->type = mobileAndZone->zoneType;
  identity->id = -1;

  for(i=0;i<zone->zonesOccupied[mobileAndZone->mobileId].nbZones;i++) {
    if (zone->zonesOccupied[mobileAndZone->mobileId].occupiedFlag[i] == ZONE_INSIDE_POLY) {
      strcpy(identity->name.name, zone->zonesMap.desc[i].name.name);
      identity->id = zone->zonesMap.desc[i].id;
      break;
    }
  }

  if (identity->id == -1)
    *bilan = S_zone_NOT_IN_TARGET;

  return ETHER;
}


/*------------------------------------------------------------------------
 * GetCurrentSpecialZone
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* zoneGetCurrentSpecialZoneExec  -  codel EXEC de GetCurrentSpecialZone
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneGetCurrentSpecialZoneExec (ZONE_MOBILE_AND_ZONE_TYPE *mobileAndZone,
					     ZONE_IDENTITY *identity, int *bilan)
{
  int i;
  ZONE_TYPE_DESC *zone;

  if (mobileAndZone->mobileId<0 || mobileAndZone->mobileId>=ZONE_NBMAX_MOBILES) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER; 
  } 

  identity->mobileId = mobileAndZone->mobileId;

  /* SPECIAL */
  zone = &zoneStructs[ZONE_SPECIAL];

  strcpy(identity->name.name, "");
  identity->type = mobileAndZone->zoneType;
  identity->id = -1;

  for(i=0;i<zone->zonesOccupied[mobileAndZone->mobileId].nbZones;i++) {
    if (zone->zonesOccupied[mobileAndZone->mobileId].occupiedFlag[i] == ZONE_INSIDE_POLY) {
      strcpy(identity->name.name, zone->zonesMap.desc[i].name.name);
      identity->id = zone->zonesMap.desc[i].id;
      break;
    }
  }

  if (identity->id == -1)
    *bilan = S_zone_NOT_IN_SPECIAL;

  return ETHER;
}

/*------------------------------------------------------------------------
 * IsInZone
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_zone_NOT_IN_SPECIAL
 */

/* zoneIsInZoneExec  -  codel EXEC de IsInZone
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT zoneIsInZoneExec (ZONE_TEST_INPUT *zoneIdentityInput, 
				 ZONE_TEST_OUTPUT * res, int *bilan)
{
  int i;
  ZONE_TYPE_DESC *zone;
  int zoneIndex;
  ZONE_TYPE zoneType;
  int nameLength;


  if (zoneIdentityInput->mobileId<0 || 
      zoneIdentityInput->mobileId>=ZONE_NBMAX_MOBILES) {
    *bilan = S_zone_BAD_VALUE;
    return ETHER;
  }  
  
  res->mobileId = zoneIdentityInput->mobileId;



  /* XXXXX test if monitoring on */


  printf ("zone : in %s type %d id %d ? ",
	  zoneIdentityInput->zoneName.name, zoneIdentityInput->zoneType, zoneIdentityInput->zoneId);

  /* --
     -- look for zone by type ... 
     -- */
  if (zoneIdentityInput->zoneType >= 0 && zoneIdentityInput->zoneType < ZONE_NB_TYPES) {

    zone = &zoneStructs[zoneIdentityInput->zoneType];

    /* ... and prehaps by begining of name */
    nameLength = strlen(zoneIdentityInput->zoneName.name);

    for(i=0; i<zone->zonesOccupied[zoneIdentityInput->mobileId].nbZones; i++) {

      /* zone found */
      if (zone->zonesOccupied[zoneIdentityInput->mobileId].occupiedFlag[i] == ZONE_INSIDE_POLY) {

	/* name not required OR same name */
	if (!nameLength ||
	    !strncmp(zoneIdentityInput->zoneName.name, 
		     zone->zonesMap.desc[i].name.name, 
		     nameLength)) {
	  printf (" YES\n");
	  res->zoneTest = ZONE_TRUE;
	  return ETHER;	
	}
      }
    }
    printf (" no\n");
    res->zoneTest = ZONE_FALSE;
    return ETHER;
  }
  
  /* --
     -- look for zone by id 
     -- */
  if (zoneIdentityInput->zoneId != -1) {
    if (lookForZoneFromId(zoneIdentityInput->zoneId, &zoneType, &zoneIndex)) {
      
      zone = &zoneStructs[zoneType];
      if (zone->zonesOccupied[zoneIdentityInput->mobileId].occupiedFlag[zoneIndex] == ZONE_INSIDE_POLY) {
	printf (" YES\n");
	res->zoneTest = ZONE_TRUE;
      }
      else {
	printf (" no\n");
	res->zoneTest = ZONE_FALSE;
      }
      return ETHER;
    }
  }

  /* --
     -- look for zone by complete name
     -- */
  if (lookForZoneFromName(zoneIdentityInput->zoneName.name, &zoneType, &zoneIndex)) {
    zone = &zoneStructs[zoneType];
    if (zone->zonesOccupied[zoneIdentityInput->mobileId].occupiedFlag[zoneIndex] == ZONE_INSIDE_POLY) {
      printf (" YES\n");
      res->zoneTest = ZONE_TRUE;
    }
    else {
      printf (" no\n");
      res->zoneTest = ZONE_FALSE;
    }
    return ETHER;
  }

  /* zone not found */
  printf ("zoneIsInZoneExec: BAD VALUE\n");
  *bilan = S_zone_BAD_VALUE;
  return ETHER;
}


/*------------------------------------------------------------------------
 * MobileToMonitor
 *
 * Description: 
 *
 * Reports:      OK
 *              S_zone_CANNOT_READ_POSTER
 */

/* zoneMobileToMonitorExec  -  codel EXEC of MobileToMonitor
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
zoneMobileToMonitorExec(ZONE_MOBILE_DESC *mobileParams, ZONE_MOBILE_ID *mobileId, int *report)
{
  int i, j;

  /* look if already existing mobile */
  for (i=0; i<ZONE_NBMAX_MOBILES; i++) {
    if (mobiles[i].pomPosterId) {
      if (strncmp(mobileParams->posPosterName.name, mobiles[i].posPosterName.name, 
		  sizeof(mobileParams->posPosterName.name) == 0)) {
	break;
      }
    }
  }

  /* look for free mobile */
  if (i==ZONE_NBMAX_MOBILES) {
    for (i=0; i<ZONE_NBMAX_MOBILES; i++) {
      if (mobiles[i].pomPosterId==NULL) {
	break;
      }
    }
  }

  /* Too many mobiles */
  if (i==ZONE_NBMAX_MOBILES) {
    *report = S_zone_TOO_MANY_MOBILES;
    *mobileId = -1;
    return ETHER;
  }

  /* (re)-record the mobile */
  *mobileId = i;
  strcpy(mobiles[i].posPosterName.name, mobileParams->posPosterName.name);
   
  /* look for its pos poster */
  if (!zoneFindPoster(mobileParams->posPosterName.name, 
		      &mobiles[i].pomPosterId,
		      &mobiles[i].pomPosterStructType)) {
    *mobileId = -1;
    *report = S_zone_CANNOT_READ_POSTER;
    return ETHER;
  }

  /* record type of monitoring */
  for (j=0; j<ZONE_NB_TYPES; j++)
    mobiles[i].monitorZoneType[j] = mobileParams->monitorZoneType[j];

  return ETHER;
}

/*------------------------------------------------------------------------
 * zoneFindPoster
 *
 * Description: 
 *
 */
static BOOL zoneFindPoster(char *name, POSTER_ID *posterId, 
			   ZONE_POSTER_POS_TYPE *posterStructType) 
{
  int length;
  
  /* Set the pomPosterId */
  if (posterFind(name, posterId) != OK) {
    printf ("zone: posterFind(%s) failed\n", name);
    return FALSE;
  }
  if (posterIoctl(*posterId, FIO_GETSIZE, &length) == ERROR) {
    printf ("zone: posterIoctl(%s) failed\n", name);
    return FALSE;
  }
  if (length == sizeof(POM_POS)) {
    *posterStructType = ZONE_POSTER_POM_POS;
    return TRUE;
  }
  if (length == sizeof(POM_ME_POS)) {
    *posterStructType = ZONE_POSTER_POM_ME_POS;
    return TRUE;
  }
  if (length == sizeof(POM_SENSOR_POS)) {
    *posterStructType = ZONE_POSTER_POM_SENSOR_POS;
    return TRUE;
  }
  if (length == sizeof(POM_EULER_POS)) {
    *posterStructType = ZONE_POSTER_POM_EULER_POS;
    return TRUE;
  }
  printf ("zone: %s bad poster type (require POM_POS or POM_ME_POS or POM_SENSOR_POS or POM_EULER_POS)\n", name);
  return FALSE;
}


/*------------------------------------------------------------------------
 * RemoveMobile
 *
 * Description: 
 *
 * Reports:      OK
 *              S_zone_BAD_VALUE
 */

/* zoneRemoveMobileExec  -  codel EXEC of RemoveMobile
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
zoneRemoveMobileExec(ZONE_MOBILE_ID *mobileId, int *report)
{
  if (*mobileId>=0 && *mobileId<ZONE_NBMAX_MOBILES) {
    mobiles[*mobileId].pomPosterId = NULL;
  }
  else {
    *report = S_zone_BAD_VALUE;
  }
  return ETHER;
}
