#ifndef UNIX
#include <vxWorks.h>
#else
#include <portLib.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <h2mathLib.h>

#include "server/zoneHeader.h"
#include "server/zonePosterWriteLibProto.h"
#include "zoneStruct.h"

/* segloc module */
#include <server/seglocPosterLib.h>
#include <seglocMapIO.h>
#include <polygons.h>

/* sick module */
#include <sickStruct.h>
#include <sick_convert.h>

/* aspect module */
#include <aspectStruct.h>
#include <server/aspectPosterLibStruct.h>

/* mapkit library */
#include <aspectLib.h>

#include "zoneNavMap.h"


#define VERBOSE 1

/* XXXXXX DEBUG */
POSTER_ID *obstPosterIdEx, *prev_obstPosterIdEx;

static BOOL zoneNavMap(SEGLOC_MAP *seglocMap, ZONES_MAP *obsts);

static BOOL zoneFilterObsts(ZONE_OBSTS_FIELD *obstsField,
			    double distSensorRobot,
			    int nbAspectSegs,
			    SICK_SEG * segsCart,
			    ASPECT_POLAR_SEG * segsPolar,
			    int *segsid,
			    ZONE_NAVOBSTS_POSTER_STR *obsts);

static void zonePrintObsts(ZONE_NAVOBSTS_POSTER_STR *obsts);
static void zoneObstPos(ZONE_OBST *obst, SICK_SEG *segInS,
			ASPECT_CONFIG *robInSensor);

static ZONE_NAVMAP_POSTER_STR posterNavMap;

static ASPECT_SEG *aspectInputMapSegs = NULL;
static int *aspectSegsid = NULL;

static int nbInputDynSegs;          /* nb segs from sick */
static int nbInputMapSegs;          /* nb segs from map */
static int nbInputSegs;             /* nb segs currently in the input map */
static int nbMaxInputSegs;          /* input map capacity */

static float robotDiameter;

void zoneSetRobotDiameter(float diameter)
{ robotDiameter = robotDiameter; }

/*------------------------------------------------------------------------
 *
 * zoneSaveNavMap  -  
 *
 * Description: 
 * 
 * Retourne:    TRUE or FALSE (if pb)
 */

BOOL zoneSaveNavMap(char *filename)
{
  char filenameMap[64];
  sprintf(filenameMap, "%s.map", filename);

  if (!seglocWriteSeglocMap (filenameMap, 
			     &posterNavMap.Map.header,
			     posterNavMap.Map.segs))
    return FALSE;

  return TRUE;
}

/*------------------------------------------------------------------------
 *
 * zoneExportNavMap  -  computes nav map from segloc map and virt obst
 *                      (removes segs that are within obsts)
 *                      and exports it in NavMap poster
 * Description: 
 * 
 * Retourne:    TRUE or FALSE (if pb)
 */

BOOL zoneExportNavMap(ZONE_SEGLOC_MAP *whereSeglocMap, ZONES_MAP *obsts,
		      int *bilan)
{
  SEGLOC_MAP *seglocMap = &posterNavMap.Map;
  POSTER_ID posterId;

  
  /* look for segloc map from poster */
  if (whereSeglocMap->source == ZONE_FROM_POSTER) {
    if (posterFind(whereSeglocMap->name.name, &posterId) == ERROR) {
      *bilan = S_zone_CANNOT_READ_POSTER;
      return FALSE;
    }

    if (zoneSEGLOC_MAPPosterRead(posterId, seglocMap) == ERROR) {
      printf ("zoneExportNavMap: failed read poster %s\n", 
	      whereSeglocMap->name.name);
      *bilan = S_zone_CANNOT_READ_POSTER;
      return FALSE;
    }
  }

  /* look for segloc map by file */
  else {
    seglocMap->header.envName[0] = 0;
    if (!seglocReadSegkit2SeglocMap(whereSeglocMap->name.name, seglocMap)) {
      printf ("zoneExportNavMap: failed read file %s\n", 
	      whereSeglocMap->name.name);
      return FALSE;
    }
  }

  /* computes new map and export in poster (WARNING: modifies seglocMap) */
  zoneNavMap(seglocMap, obsts);

  /* export in poster */
  if (zoneNavMapPosterWrite(ZONE_NAVMAP_POSTER_ID, &posterNavMap) != OK) {
    printf ("zoneNavMap: zoneNavMapPosterWrite failed\n");
    return FALSE;
  }
  
  /* record in local data */
  if (nbMaxInputSegs 
      < posterNavMap.Map.header.nbSegs + SICK_MAX_ACQ_SEGS) {
    if (aspectInputMapSegs) {
      free(aspectInputMapSegs);
      aspectInputMapSegs =  NULL;
      free(aspectSegsid);
    }
  }
  
  if (!aspectInputMapSegs) {
    nbMaxInputSegs = posterNavMap.Map.header.nbSegs + SICK_MAX_ACQ_SEGS;
    if (!(aspectInputMapSegs = malloc(nbMaxInputSegs*sizeof(SICK_SEG)))) {
      *bilan = S_zone_NOT_ENOUGH_MEMORY;
      return FALSE;
    }
    if (!(aspectSegsid = malloc(nbMaxInputSegs*sizeof(int)))) {
      *bilan = S_zone_NOT_ENOUGH_MEMORY;
      return FALSE;
    }
  }
 
  /* navMapSegs */
  memcpy(aspectInputMapSegs, posterNavMap.Map.segs,   
	 posterNavMap.Map.header.nbSegs*sizeof(SICK_SEG));
  nbInputMapSegs = posterNavMap.Map.header.nbSegs;
  nbInputSegs = nbInputMapSegs;

  return TRUE;
}

/*------------------------------------------------------------------------
 *
 * zoneNavMap  - remove from seglocMap the segs that are within virt obsts
 *               and adds obsts borders
 *               
 *               WARNING : modifies seglocMap !
 * 
 * Retourne:    TRUE or FALSE
 */

static BOOL zoneNavMap(SEGLOC_MAP *seglocMap, ZONES_MAP *obsts)
{
  int seg;
  int obst;
  int pt;
  int nbNavSegs=0;
  int firstPt;
  BOOL inside;
  double segPts[4];
  SEGLOC_MAP *navMap=seglocMap;

  /* for each seg */
  for (seg=0; seg<seglocMap->header.nbSegs; seg++) {

    inside = FALSE;

    /* for each virtual obst */
    firstPt = 0;
    for (obst = 0; obst < obsts->nbZones; obst++) {
    
      segPts[0] = seglocMap->segs[seg].x1;
      segPts[1] = seglocMap->segs[seg].y1;
      segPts[2] = seglocMap->segs[seg].x2;
      segPts[3] = seglocMap->segs[seg].y2;

      /* the seg belongs to an obst : forget it */
      if (PLpolygonInsidePolygon(2, 
				 segPts,
				 obsts->nbPtsPerZone[obst], 
				 (double *)(&obsts->points[firstPt].x))) {
	inside = TRUE;
	break;
      }
      firstPt += obsts->nbPtsPerZone[obst];
    }
    
    /* record the seg */
    if (!inside) {
      navMap->segs[nbNavSegs] = seglocMap->segs[seg];
      nbNavSegs++;
    }
  } /* for each map seg */

  /* record the obst borders */
  firstPt = 0;
  for (obst = 0; obst < obsts->nbZones; obst++) {

    for (pt=firstPt; pt<firstPt+obsts->nbPtsPerZone[obst]-1; pt++) {
      navMap->segs[nbNavSegs].x1 = obsts->points[pt].x;
      navMap->segs[nbNavSegs].y1 = obsts->points[pt].y;
      navMap->segs[nbNavSegs].x2 = obsts->points[pt+1].x;
      navMap->segs[nbNavSegs].y2 = obsts->points[pt+1].y;
      nbNavSegs++;
    }
    navMap->segs[nbNavSegs].x1 = obsts->points[pt].x;
    navMap->segs[nbNavSegs].y1 = obsts->points[pt].y;
    navMap->segs[nbNavSegs].x2 = obsts->points[firstPt].x;
    navMap->segs[nbNavSegs].y2 = obsts->points[firstPt].y;
    nbNavSegs++;
    firstPt += obsts->nbPtsPerZone[obst];
  } /* obst */
  
  navMap->header.nbSegs = nbNavSegs;

  printf ("zoneNavMap: %d segs recorded\n", nbNavSegs);

  return TRUE;
}


/*------------------------------------------------------------------------
 *
 * zoneObstsInNav  -  lists of the obstacle that are in navigation space
 *               and visible from robot view point. Exports result in
 *               poster NavObsts
 * 
 * Retourne:    TRUE or FALSE
 */

BOOL zoneObstsInNav(ZONE_OBSTS_FIELD *obstsField, 
		    int *bilan)
{
  static SICK_SEGS_POSTER_STR sickSegs;
  ASPECT_CONFIG viewPointInMap;
  ASPECT_VIEW_PARAMETERS viewParameters;
  int nbAspectSegs;
  double distSensorRobot;

  static ASPECT_CARTESIANASPECT_POSTER_STR aspectCartesianMap;
  static ASPECT_POLARASPECT_POSTER_STR aspectPolarMap;
  static ZONE_NAVOBSTS_POSTER_STR obsts;

  static POSTER_ID obstPosterId, prev_obstPosterId;

  /* test parameters */
  if (obstsField->spreadAngle <= 0.01) {
    printf ("zoneObstsInNav: bad parameter for spreadAngle (%g)\n",
	    obstsField->spreadAngle);
    *bilan = S_zone_BAD_VALUE;
    return FALSE;
  }

  /* test if we have a nav map */
  if (!aspectInputMapSegs) {
    printf ("zoneObstsInNav: WARNING no nav map (ie, no static obstacles)\n");
    nbMaxInputSegs = SICK_MAX_ACQ_SEGS;
    if (!(aspectInputMapSegs = malloc(nbMaxInputSegs*sizeof(SICK_SEG)))) {
      *bilan = S_zone_NOT_ENOUGH_MEMORY;
      return FALSE;
    }
    if (!(aspectSegsid = malloc(nbMaxInputSegs*sizeof(int)))) {
      *bilan = S_zone_NOT_ENOUGH_MEMORY;
      return FALSE;
    }
  }

  /* look for sick poster */
  if (!obstPosterId) {
    if (posterFind("sickSegs", &obstPosterId) == ERROR ) {
      printf ("obstPosterId %d\n", (int)obstPosterId);
      *bilan = S_zone_CANNOT_READ_POSTER;
      return FALSE;
    }
    prev_obstPosterId = obstPosterId;
    prev_obstPosterIdEx = &prev_obstPosterId;
    obstPosterIdEx = &obstPosterId;
  }

  /* read dyn segs */
  if (zoneSICK_SEGS_POSTER_STRPosterRead(obstPosterId, &sickSegs) 
      == ERROR) {
    *bilan = S_zone_CANNOT_READ_POSTER;
    return FALSE;
  }

  /* convert in global frame AND record in input map */
  nbInputSegs = nbInputMapSegs;
  sickConvertSegsFrame(&sickSegs.Header, 
		       SICK_GLOBAL_FRAME,
		       sickSegs.Segs.segs, 
		       (SICK_SEG *) &aspectInputMapSegs[nbInputSegs]);
  nbInputDynSegs = sickSegs.Header.nbSegs;
  nbInputSegs = nbInputMapSegs + nbInputDynSegs;

  /* computes virtual sensor pos in global frame */
  viewPointInMap.theta = sickSegs.Header.sickPomPos.mainToOrigin.euler.yaw +
		       obstsField->direction;
  /* move back the virtual sensor in order to entirely integrate
     the robot within the considered angular sector */
  if (obstsField->spreadAngle < M_PI) {
    distSensorRobot = robotDiameter/tan(obstsField->spreadAngle/2.0);

    /* to not be behind the obstacles of the map */
    if (distSensorRobot > robotDiameter/2.0)
      distSensorRobot = robotDiameter/2.0;
  }
  else
    distSensorRobot = 0.0;
  viewPointInMap.x = sickSegs.Header.sickPomPos.mainToOrigin.euler.x
		   - distSensorRobot * cos(viewPointInMap.theta);
		   
  viewPointInMap.y = sickSegs.Header.sickPomPos.mainToOrigin.euler.y
		   - distSensorRobot * sin(viewPointInMap.theta);

  /* aspect parameters */
  viewParameters.maxDistView = obstsField->range 
			       + robotDiameter/2.0 
			       + distSensorRobot;
  viewParameters.halfAngleView = obstsField->spreadAngle/2.0;
  viewParameters.keepReversedSegs = 1;
  viewParameters.minAspectSegAngle = 0;
  
  if (prev_obstPosterId != obstPosterId) {
    printf ("BUG IN ZONE 1 !!\n");
    obstPosterId = 0;
    prev_obstPosterId = 0;
  }
      
  /* computes aspect */
  nbAspectSegs = aspectMapFillIn(&viewPointInMap, 
				 &viewParameters,
				 /* input map */
				 nbInputSegs, 
				 aspectInputMapSegs, 
				 FALSE, /* results in local sensor frame */
				 MIN(ASPECT_MAX_ASPECT_MAP_SIZE,
				     SICK_MAX_ACQ_SEGS), /* where to fill result */
				 (ASPECT_SEG *)aspectCartesianMap.Segs.segs,
				 aspectPolarMap.PolarSegs.segs,
				 aspectSegsid,
				 /* simul laser */
				 0, NULL);

  if (prev_obstPosterId != obstPosterId) {
    printf ("BUG IN ZONE 2 !!\n");
    obstPosterId = 0;
    prev_obstPosterId = 0;
  }
      
  /* filter obsts and free space */
  zoneFilterObsts(obstsField,
		  distSensorRobot,
		  nbAspectSegs, 
		  aspectCartesianMap.Segs.segs,
		  aspectPolarMap.PolarSegs.segs,
		  aspectSegsid,
		  &obsts);

  if (prev_obstPosterId != obstPosterId) {
    printf ("BUG IN ZONE 2 !!\n");
    obstPosterId = 0;
    prev_obstPosterId = 0;
  }
      
  /* record in poster */
  if (zoneNavObstsPosterWrite(ZONE_NAVOBSTS_POSTER_ID, 
			      &obsts) == ERROR) {
    *bilan = S_zone_CANNOT_WRITE_POSTER;
    return FALSE;
  }

  /* print */
  zonePrintObsts(&obsts);

  return TRUE;
}

/*------------------------------------------------------------------------
 *
 * zonePrintObsts  -  
 *
 * Description: 
 * 
 * Retourne:    
 */


static void zonePrintObsts(ZONE_NAVOBSTS_POSTER_STR *obsts)
{
#if VERBOSE>0
  static ZONE_OBST_STATE prevState1, prevState2;
#endif

#if VERBOSE>2
  int numObst;
#endif

#if VERBOSE>0
  if (prevState1 != obsts->Header.obstState ||
      prevState2 != SDI_F->obstState) {
    printf ("%d dyn/%d  %d free : %5s  %5s (cart %d%%  pol %d%%)\n" ,
	    obsts->Header.nbDynObsts,
	    obsts->Header.nbObsts,
	    obsts->Header.nbFreeWays,
	    
	    obsts->Header.obstState == prevState1 ? "" :
	    obsts->Header.obstState == ZONE_FREE ? "free" :
	    obsts->Header.obstState == ZONE_HIGH_CLUTTERED ? "high" :
	    obsts->Header.obstState == ZONE_LOW_CLUTTERED ? "low" : "block",
	    
	    SDI_F->obstState == prevState2 ? "" :
	    SDI_F->obstState == ZONE_FREE ? "FREE" :
	    SDI_F->obstState == ZONE_HIGH_CLUTTERED ? "HIGH" :
	    SDI_F->obstState == ZONE_LOW_CLUTTERED ? "LOW" : "BLOCK",
	    
	    obsts->Header.clutteredPercentCart,
	    obsts->Header.clutteredPercentPolar);
    prevState1 = obsts->Header.obstState;
    prevState2 = SDI_F->obstState;
  }
#endif

#if VERBOSE>1
  printf (" obst sum  cart %5.2fm  polar %5.2fdeg\n"
	  " free sum  cart %5.2fm  polar %5.2fdeg\n",
	  obsts->Header.obstsLengthSumCart , 180/M_PI*obsts->Header.obstsLengthSumPolar,
	  obsts->Header.freeWaysLengthSumCart , 180/M_PI*obsts->Header.freeWaysLengthSumPolar);
#endif
  
#if VERBOSE>1
  printf (" %d/100 clutteredC  / %d/100 clutteredP\n",
	  obsts->Header.clutteredPercentCart,
	  obsts->Header.clutteredPercentPolar);
#endif

#if VERBOSE>2
  for(numObst=0; numObst<obsts->Header.nbDynObsts; numObst++) {
    printf ("   dyn %d: th = %5.2f deg  dist = %5.2f m   width = %5.2f m\n",
	    numObst, 
	    180/M_PI*obsts->Obstacles[numObst].angleMean,
	    obsts->Obstacles[numObst].distMin,
	    obsts->Obstacles[numObst].width);
  }
#endif  
}

/*------------------------------------------------------------------------
 *
 * zoneFilterObsts  -  
 *
 * Description: 
 * 
 * Retourne:    
 */

static BOOL zoneFilterObsts(ZONE_OBSTS_FIELD *obstsField,
			    double distSensorRobot,
			    int nbAspectSegs,
			    SICK_SEG * segsCart,
			    ASPECT_POLAR_SEG * segsPolar,
			    int *segsid,
			    ZONE_NAVOBSTS_POSTER_STR *obsts)
{
  double r0, t0, x0, y0;
  double r1, t1, x1, y1;
  double r2, t2, x2, y2;
  int nbBackObst;
  double widthFree, angleFree, distMin;
  int i, numObst, prevObst=-1;
  SICK_SEG segInS;
  ASPECT_CONFIG robInSensor;
  double theoricalFreeWaysLengthSumCart;
  double theoricalFreeWaysLengthSumPolar;
  ZONE_OBST_TYPE type;
  static int count;

  /* reset obstacles */
  obsts->Header.id++;
  obsts->Header.nbObsts = 0;
  obsts->Header.nbDynObsts = 0;  /* number of non-map obstacles (ie, dyn) */
  obsts->Header.nbFreeWays = 0;  /* number of free ways large enough for the robot */
  obsts->Header.obstsLengthSumCart = 0;
  obsts->Header.freeWaysLengthSumCart = 0;
  obsts->Header.obstsLengthSumPolar = 0;
  obsts->Header.freeWaysLengthSumPolar = 0;
  nbBackObst = 0;

  robInSensor.x = distSensorRobot;
  robInSensor.y = 0;
  robInSensor.theta = -obstsField->direction;

  /* record obstacles */
  r0 = segsPolar[0].r1;
  t0 = - obstsField->spreadAngle/2.0;
  x0 = r0 * cos(t0);
  y0 = r0 * sin(t0);

  for (i=0; i<nbAspectSegs; i++) {

    /* filter backward (ie, map) obstacles */
    distMin = MIN(segsPolar[i].r1,
		  segsPolar[i].r2);
    if (distMin < distSensorRobot) {
      nbBackObst++;
      continue;
    }

    /* free zone t0 -> t1 */
    t1 = segsPolar[i].t1;
    r1 = segsPolar[i].r1;
    x1 = segsCart[i].x1;
    y1 = segsCart[i].y1;

    angleFree = fabs(angleLimit(t1 - t0));
    widthFree = hypot(x1-x0, y1-y0);
    if (widthFree > robotDiameter) {
      obsts->Header.freeWaysLengthSumCart += widthFree;
      obsts->Header.freeWaysLengthSumPolar += angleFree;
      numObst = nbAspectSegs + obsts->Header.nbFreeWays;
      if (numObst >= ZONE_NB_MAX_OBSTS) {
	printf ("zoneFilterObsts: too many obst. Free zone %d removed\n",
		obsts->Header.nbFreeWays);
      }
      else {
	obsts->Obstacles.obst[numObst].type = ZONE_FREE_WAY;

	segInS.x1 = x0;
	segInS.y1 = y0;
	segInS.x2 = x1;
	segInS.y2 = y1;
	zoneObstPos(&obsts->Obstacles.obst[numObst], &segInS, &robInSensor);

	obsts->Header.nbFreeWays++;
      }
    }
    
    /* obst t1 -> t2 */
    t2 = segsPolar[i].t2;
    r2 = segsPolar[i].r2;
    x2 = segsCart[i].x2;
    y2 = segsCart[i].y2;
    
    /* dyn seg */
    if (segsid[i] >= nbInputMapSegs) {

      /* merge with previous dyn obstacles if any */
      if (widthFree < 0.02 
	  && prevObst != -1
	  && obsts->Obstacles.obst[prevObst].type == ZONE_DYN_OBST) {
	t1 = segsPolar[i-1].t1;
	r1 = segsPolar[i-1].r1;
	x1 = segsCart[i-1].x1;
	y1 = segsCart[i-1].y1;
#if VERBOSE>1
	printf("fuse dyn obsts\n");
#endif
	nbBackObst++;
      }
      /* or new obst */
      else {
	obsts->Header.nbObsts++;
	obsts->Header.nbDynObsts++;
      }
      numObst = obsts->Header.nbDynObsts-1;
      type = ZONE_DYN_OBST;
    }

    /* map seg */
    else {
      obsts->Header.nbObsts++;
      numObst = nbAspectSegs - (obsts->Header.nbObsts - obsts->Header.nbDynObsts);
      type = ZONE_STATIC_OBST;
    }

    if (numObst >= ZONE_NB_MAX_OBSTS) {
      printf ("zoneFilterObsts: too many obst. obst %d removed\n",
	      numObst);
      obsts->Header.nbObsts--;
      if (type == ZONE_DYN_OBST)
	obsts->Header.nbDynObsts--;
    }
    else {
      obsts->Obstacles.obst[numObst].type = type;
      prevObst = numObst;

      segInS.x1 = x1;
      segInS.y1 = y1;
      segInS.x2 = x2;
      segInS.y2 = y2;
      zoneObstPos(&obsts->Obstacles.obst[numObst], &segInS, &robInSensor);
      
      obsts->Header.obstsLengthSumCart += obsts->Obstacles.obst[numObst].width;
      obsts->Header.obstsLengthSumPolar += fabs(angleLimit(t2 - t1));
      
      r0 = r2;
      t0 = t2;
      x0 = x2;
      y0 = y2;
    }
  }  /* for */

  /* last free space */
  r1 = r2;
  t1 = obstsField->spreadAngle/2.0;
  x1 = r1 * cos(t1);
  y1 = r1 * sin(t1);

  angleFree = fabs(angleLimit(t1 - t0));
  widthFree = hypot(x1-x0, y1-y0);
  if (widthFree > robotDiameter) {
    obsts->Header.freeWaysLengthSumCart += widthFree;
    obsts->Header.freeWaysLengthSumPolar += angleFree;
    numObst = nbAspectSegs + obsts->Header.nbFreeWays;
    if (numObst >= ZONE_NB_MAX_OBSTS) {
      printf ("zoneFilterObsts: too many obst. Last free zone %d removed\n",
	      obsts->Header.nbFreeWays);
    }
    else {
      obsts->Obstacles.obst[numObst].type = ZONE_FREE_WAY;

      segInS.x1 = x0;
      segInS.y1 = y0;
      segInS.x2 = x1;
      segInS.y2 = y1;
      zoneObstPos(&obsts->Obstacles.obst[numObst], &segInS, &robInSensor);
      obsts->Header.nbFreeWays++;
    }
  }

  /* clustered measurement in cartesian space */
  theoricalFreeWaysLengthSumCart = obsts->Header.freeWaysLengthSumCart +
				   obsts->Header.obstsLengthSumCart;
  if (theoricalFreeWaysLengthSumCart == 0)
    obsts->Header.clutteredPercentCart = 0;
  else if (!obsts->Header.nbFreeWays)
    obsts->Header.clutteredPercentCart = 100;
  else
    obsts->Header.clutteredPercentCart = (int)(100*(1- obsts->Header.freeWaysLengthSumCart/
					    theoricalFreeWaysLengthSumCart));

  /* clustered measurement in polar space */
  theoricalFreeWaysLengthSumPolar = obsts->Header.freeWaysLengthSumPolar +
				    obsts->Header.obstsLengthSumPolar;
  if (theoricalFreeWaysLengthSumPolar == 0)
    obsts->Header.clutteredPercentPolar = 0;
  else if (!obsts->Header.nbFreeWays)
    obsts->Header.clutteredPercentCart = 100;
  else
    obsts->Header.clutteredPercentPolar = (int)(100*(1- obsts->Header.freeWaysLengthSumPolar/
					     theoricalFreeWaysLengthSumPolar));

  if (obsts->Header.clutteredPercentCart == 100)
    obsts->Header.obstState = ZONE_BLOCKED;
  else if (obsts->Header.clutteredPercentCart > 60)
    obsts->Header.obstState = ZONE_HIGH_CLUTTERED;
  else if (obsts->Header.clutteredPercentCart > 20)
    obsts->Header.obstState = ZONE_LOW_CLUTTERED;
  else
    obsts->Header.obstState = ZONE_FREE;

  /* record the new state */
  if (SDI_F->obstState != obsts->Header.obstState) {

    /* wait a little before indicating a block */
    if (obsts->Header.obstState == ZONE_FREE ||
	obsts->Header.obstState == ZONE_LOW_CLUTTERED ||
	count++ == 10) {
      SDI_F->obstState = obsts->Header.obstState;
      count = 0;
    }
  }

  /* shift arrays in case of back obsts that have been removed */
  if (nbBackObst) {
    for (i=0; i<obsts->Header.nbDynObsts + obsts->Header.nbFreeWays; i++) {
      obsts->Obstacles.obst[obsts->Header.nbObsts+i] = obsts->Obstacles.obst[obsts->Header.nbObsts+nbBackObst+i];
    }
  }

  return TRUE;
}

/*------------------------------------------------------------------------
 *
 * zoneObstPos  -  
 *
 * Description: 
 * 
 * Retourne:    
 */

static void zoneObstPos(ZONE_OBST *obst, SICK_SEG *segInS,
			ASPECT_CONFIG *robInSensor) 
{
  double crs, srs;
  SICK_SEG segInR;
  
  /* sensor frame -> robot frame */
  crs = cos(robInSensor->theta);
  srs = sin(robInSensor->theta);
  
  segInR.x1 = (segInS->x1 - robInSensor->x)* crs + segInS->y1 * srs;
  segInR.y1 = (robInSensor->x - segInS->x1)* srs + segInS->y1 * crs;
  
  segInR.x2 = (segInS->x2 - robInSensor->x)* crs + segInS->y2 * srs;
  segInR.y2 = (robInSensor->x - segInS->x2)* srs + segInS->y2 * crs;
  
  /* obst params */
  obst->angleMean = (atan2(segInR.y1,segInR.x1)
		     + atan2(segInR.y2,segInR.x2))/2.0;
  obst->distMin = MIN(hypot(segInR.y1,segInR.x1), 
		      hypot(segInR.y2,segInR.x2)) 
		  - robotDiameter/2.0;
  obst->width = hypot(segInR.y2-segInR.y1, 
		      segInR.x2-segInR.x1);
}
