/**
 ** zoneIO.c
 **
 ** Fonctions pour manipuler les structures ZONES
 **
 ** Auteur: Sara FLEURY
 ** Date: 11/15/01
 **
 **/


#ifdef VXWORKS
#include <vxWorks.h>
#else
#include <portLib.h>
#endif
#include <stdio.h>
#include <string.h>
#include <h2mathLib.h>

#include "pomStruct.h"
#include "genBasicStruct.h"
#include "sickStruct.h"
#include "zoneConst.h"
#include "zoneStruct.h"
#include "zoneIO.h"

 
/*------------------------------------------------------------------------
 * zoneReadFile
 *
 * Description: Reads zones description from a file.
 *              Fills the structure zones
 *
 * Returns:    BOOL (TRUE or FALSE)
 */

BOOL
zoneReadFile(char *filename, ZONES_MAP *zonesMap)
{
  FILE *file;
  int i = 0;
  int j, nbPts;
  char line[1024];
  char class[64];
  char *newline;

  /* init zonesPoly et zonesDesc */
  zonesMap->nbZones = 0;
  zonesMap->nbTotalPoints = 0;

  /* Open file */
  if ((file = fopen(filename, "r")) == NULL) {
    printf ("zoneReadFileExec: cannot read file %s\n", filename);
    return FALSE;
  }
  
  /* Get next line */
  while(fgets(line,1024,file)) {
    
    if (strncmp("name: ",line,sizeof("name"))!=0) 
      continue;

    /* start a new zone */
    if (i >= ZONE_NBMAX_ZONES) {
      printf ("zoneReadFileExec: too many zones (max=%d)\n", 
	      ZONE_NBMAX_ZONES);
      break;
    }
    if ((newline = strstr(line,"\n")))
	*newline = '\0';
    else {
	printf ("no new line\n");
    }
    bzero(zonesMap->desc[i].name.name, sizeof(GEN_STRING32));
    strncpy(zonesMap->desc[i].name.name,
	    &line[sizeof("name:")], sizeof(line) - sizeof("name:"));
    printf ("name[%d]: \"%s\"\n", i, zonesMap->desc[i].name.name);

    /* Read class */
    if (!fgets(line,1024,file)) {
      printf("zoneReadFileExec: stop\n");
      break;
    }
    if (sscanf(line, "class: %s\n", class) != 1) {
      printf ("zoneReadFileExec: cannot read class zone \"%s\"\n",
	      zonesMap->desc[i].name.name);
      break;
    }
    if (strcmp(class, "ZONE_TOPO") == 0)
      zonesMap->desc[i].class = ZONE_TOPO;
    else if (strcmp(class, "ZONE_STRESS") == 0)
      zonesMap->desc[i].class = ZONE_STRESS;
    else {
      printf("zoneReadFileExec: unknown class %s. Stop.\n", class);
      break;
    }
    printf ("   class: %s\n", zoneClassName(zonesMap->desc[i].class));
    
    /* Read nb pts */
    if (!fgets(line,1024,file)) {
      printf("zoneReadFileExec: stop\n");
      break;
    }
    if (sscanf(line, "points: %d\n", &nbPts) != 1) {
      printf ("zoneReadFileExec: cannot read nb points zone \"%s\"\n",
	      zonesMap->desc[i].name.name);
      break;
    }
    if (nbPts + zonesMap->nbTotalPoints > ZONE_NBMAX_PTS) {
      printf ("zoneReadFileExec: too many pts (max=%d)\n", 
	      ZONE_NBMAX_PTS);
      break;
    }
    
    /* Read points */
    for(j=0; j<nbPts; j++) {

      if (!fgets(line,1024,file)) {
	printf("zoneReadFileExec: stop\n");
	break;
      }
      if (sscanf(line, "%lf %lf\n", 
		 &zonesMap->points[zonesMap->nbTotalPoints +j].x,
		 &zonesMap->points[zonesMap->nbTotalPoints +j].y) != 2) {
	printf ("zoneReadFileExec: cannot read point %d zone \"%s\"\n"
		"                  keep only %d points (instead of %d)\n",
		j, zonesMap->desc[i].name.name, j-1, nbPts);
	nbPts = j-1;
	break;
      }
      
    }
    
    /* test if first and last pt identical */
    if (zonesMap->points[zonesMap->nbTotalPoints +j-1].x == 
	zonesMap->points[zonesMap->nbTotalPoints].x &&
	zonesMap->points[zonesMap->nbTotalPoints +j-1].y == 
	zonesMap->points[zonesMap->nbTotalPoints].y) {
      printf ("zoneReadFileExec : remove last pt identical to first in %s\n",
	      zonesMap->desc[i].name.name);
      nbPts--;
    }

    /* update points number */
    zonesMap->nbTotalPoints += nbPts;
    zonesMap->nbPtsPerZone[i] = nbPts;
    printf ("   nbPts: %d \n", zonesMap->nbPtsPerZone[i]);


    /* read final pos */
    if (!fgets(line,1024,file)) {
	printf("zoneReadFileExec: stop\n");
	break;
    }
    if (sscanf(line, "FinalPos: %lf %lf %lf\n", 
	       &zonesMap->desc[i].finalPos.x, 
	       &zonesMap->desc[i].finalPos.y,
	       &zonesMap->desc[i].finalPos.theta) != 3) {
      printf ("zoneReadFileExec: cannot read finalPos zone \"%s\"\n",
	      zonesMap->desc[i].name.name);
      break;
    }

    /* read target pos */
    if (!fgets(line,1024,file)) {
	printf("zoneReadFileExec: stop\n");
	break;
    }
    if (sscanf(line, "Target: %lf %lf %lf\n", 
	       &zonesMap->desc[i].target.x, 
	       &zonesMap->desc[i].target.y,
	       &zonesMap->desc[i].target.theta) != 3) {
      printf ("zoneReadFileExec: cannot read targetPos zone \"%s\"\n",
	      zonesMap->desc[i].name.name);
      break;
    }


    /* Record this zone */
    zonesMap->nbZones++;
      
    
    /* Next zone */
    i++;
  }

  fclose(file);
  return TRUE;
}

/*------------------------------------------------------------------------
 * zoneSaveFile
 *
 * Description: Saves zones description in file.
 *
 * Returns:    BOOL (TRUE or FALSE)
 */

BOOL
zoneSaveFile(char *filename, ZONES_MAP *zonesMap, ZONE_POINT pt[])
{
  FILE *file;
  int i, j, firstPt=0;

  /* Open file */
  if ((file = fopen(filename, "w")) == NULL) {
    printf ("zoneSaveFileExec: cannot write file %s\n", filename);
    return FALSE;
  }
  fprintf (file, 
	   "/* --------------------------------------------------------------------- */\n"
	   "zones classs :      topo                   stress (enlarged)      "
	   "\n"
	   "             ZONE_TOPO_MANIP   \n"
	   "             ZONE_TOPO_DOCK    \n"
	   "                                         ZONE_STRESS_VIRTUAL_OBST  \n "
	   "/* ---------------------------------------------------------------------- */\n");
  
  for (i=0; i<zonesMap->nbZones; i++) {
    if (zonesMap->nbPtsPerZone[i] <= 0)
      continue;
    fprintf (file, "\nname: %s\n",
	     zonesMap->desc[i].name.name);
    switch(zonesMap->desc[i].class) {
    case ZONE_TOPO:
      fprintf (file, "class: %s\n",  "ZONE_TOPO");
      break;   
    case ZONE_STRESS:
      fprintf (file, "class: %s\n",  "ZONE_STRESS");
      break;
    }
    fprintf (file, "points: %d\n", 
	     zonesMap->nbPtsPerZone[i]);
    for (j=0; j<zonesMap->nbPtsPerZone[i]; j++) {
      fprintf (file, "%5.2f %5.2f\n", 
	       pt[firstPt+j].x,
	       pt[firstPt+j].y);
    }
    firstPt += zonesMap->nbPtsPerZone[i];
    fprintf(file, "FinalPos: %5.2f %5.2f %5.2f\n", 
	    zonesMap->desc[i].finalPos.x,
	    zonesMap->desc[i].finalPos.y,
	    zonesMap->desc[i].finalPos.theta);
    fprintf(file, "Target: %5.2f %5.2f %5.2f\n", 
	    zonesMap->desc[i].target.x,
	    zonesMap->desc[i].target.y,
	    zonesMap->desc[i].target.theta);
    
    printf ("ZONE %d \"%s\" recorded\n", 
	    i, zonesMap->desc[i].name.name);
  }
  
  fclose(file);
  return TRUE;
}


char * zoneTypeName(ZONE_TYPE type)
{
  static char *typeName[ZONE_NB_TYPES] = ZONE_TYPE_NAMES;
  return typeName[type];
}

char * zoneClassName(ZONE_CLASS class)
{
  static char *className[ZONE_NB_CLASS] = ZONE_CLASS_NAMES;
  return className[class];
}

